import unittest

import joule.api
import subprocess
from joule import api


class TestBasicAPI(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node1 = api.get_node("node1.joule")
        followers = await self.node1.follower_list()
        self.assertEqual(len(followers), 1)
        self.node2 = followers[0]
        # add the root user as a follower to node2.joule
        user = await self.node2.master_add("user", "cli")
        subprocess.run(f"joule node add node2.joule https://node2.joule:8088 {user.key}".split(" "))
        """
        node1.joule:
        /main/folder/added:int32[x]
        node2.joule:
        /main/folder/base:int32[x]
        """

    async def asyncTearDown(self):
        await self.node2.master_delete("user", "cli")
        await self.node1.close()
        await self.node2.close()

    async def test_retrieves_streams(self):
        logs = await self.node1.module_logs("Remote")
        print(logs)
        print('----------')
        added_stream = await self.node1.data_stream_get("/main/folder/added")
        base_stream = await self.node2.data_stream_get("/main/folder/base")
        # make sure added_stream has data
        added_data = await self.node1.data_subscribe(added_stream)
        base_data = await self.node2.data_subscribe(base_stream)
        base_block = await base_data.read()
        added_block = await added_data.read()
        # align the time stamps
        diff = base_block['timestamp'][0] - added_block['timestamp'][0]
        print("the diff is %d" % diff)
        await added_data.close()
        await base_data.close()

    async def test_copies_events(self):
        events = []
        for i in range(1000, 2000, 100):
            events.append(joule.api.Event(start_time=i, end_time=i + 50,
                                          content={'name': f"Event{i}"}))
        event_stream = joule.api.EventStream("TestEvents")
        event_stream = await self.node1.event_stream_create(event_stream, "/eventstreams")
        await self.node1.event_stream_write(event_stream, events)
        cmd = f"/usr/local/bin/joule event copy /eventstreams/TestEvents /eventstreams/TestEventsCopy -d node2.joule"
        subprocess.run(cmd.split(" "))
        copied_events = await self.node2.event_stream_read("/eventstreams/TestEventsCopy")
        self.assertListEqual(events, copied_events)



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp
import numpy as np
from sqlalchemy.orm import Session
import asyncio
from joule.models import DataStream, pipes
import joule.controllers
from tests.controllers.helpers import create_db, MockStore, MockSupervisor
from tests import helpers


class TestDataController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app["db"].close()
        self.app["psql"].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app["db"], app["psql"] = create_db(["/folder1/stream1:float32[x, y, z]",
                                            "/folder2/deeper/stream2:int8[val1, val2]"])
        app["data-store"] = MockStore()
        self.supervisor = MockSupervisor()
        app["supervisor"] = self.supervisor
        return app


    async def test_read_binary_data(self):
        db: Session = self.app["db"]
        store: MockStore = self.app["data-store"]
        nchunks = 10
        store.configure_extract(nchunks)
        resp: aiohttp.ClientResponse = await \
            self.client.get("/data", params={"path": "/folder1/stream1"})
        rx_chunks = 0
        async for data, _ in resp.content.iter_chunks():
            if len(data) > 0:
                rx_chunks += 1
        self.assertEqual(nchunks, rx_chunks)

        # can retrieve stream by id and as decimated data
        stream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.get("/data", params={"id": stream.id, 'decimation-level': 16})
        rx_chunks = 0
        async for data, _ in resp.content.iter_chunks():
            if len(data) > 0:
                rx_chunks += 1
        self.assertEqual(nchunks, rx_chunks)
        self.assertEqual(resp.headers['joule-decimation'], '16')
        self.assertEqual(resp.headers['joule-layout'], stream.decimated_layout)


    async def test_read_json_data(self):
        db: Session = self.app["db"]
        store: MockStore = self.app["data-store"]
        n_chunks = 10
        n_intervals = 2
        # mock data has 2 intervals retrieved as 10 chunks each
        store.configure_extract(n_chunks, n_intervals)
        resp: aiohttp.ClientResponse = await \
            self.client.get("/data.json", params={"path": "/folder1/stream1"})
        data = await resp.json()
        self.assertEqual(1, data['decimation_factor'])
        self.assertEqual(len(data['data'][0]), n_chunks * 25)
        # can retrieve stream by id and as decimated data as well
        stream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.get("/data.json", params={"id": stream.id, 'decimation-level': 16})
        data = await resp.json()
        self.assertEqual(data['decimation_factor'], 16)
        # only the interval is exposed, the chunking is transparent
        self.assertEqual(len(data['data']), 2)
        for interval in data['data']:
            self.assertEqual(len(interval), n_chunks * 25)


    async def test_subscribes_to_data(self):
        db: Session = self.app["db"]
        my_stream = db.query(DataStream).filter_by(name="stream1").one()
        blk1 = helpers.create_data(my_stream.layout)
        blk2 = helpers.create_data(my_stream.layout, length=50)
        my_pipe = pipes.LocalPipe(my_stream.layout)
        my_pipe.write_nowait(blk1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(blk2)
        my_pipe.close_nowait()
        self.supervisor.subscription_pipe = my_pipe
        async with self.client.get("/data", params={"id": my_stream.id,
                                                    "subscribe": '1'}) as resp:
            pipe = pipes.InputPipe(stream=my_stream, reader=resp.content)
            rx_blk1 = await pipe.read()
            pipe.consume(len(rx_blk1))
            np.testing.assert_array_equal(blk1, rx_blk1)
            self.assertTrue(pipe.end_of_interval)

            rx_blk2 = await pipe.read()
            pipe.consume(len(rx_blk2))
            np.testing.assert_array_equal(blk2, rx_blk2)
            with self.assertRaises(pipes.EmptyPipe):
                await pipe.read()
        self.assertEqual(self.supervisor.unsubscribe_calls, 1)


    async def test_unsubscribes_terminated_connections(self):
        db: Session = self.app["db"]
        supervisor: MockSupervisor = self.app["supervisor"]
        supervisor.hang_pipe = True
        my_stream = db.query(DataStream).filter_by(name="stream1").one()
        my_pipe = pipes.LocalPipe(my_stream.layout)
        my_pipe.close_nowait()
        self.supervisor.subscription_pipe = my_pipe
        async with self.client.get("/data", params={"id": my_stream.id,
                                                    "subscribe": '1'}):
            # ignore the data
            pass
        # NOTE: unsubscribe is called but the exception propogates and somehow
        # we loose the reference so this assert fails. Not a problem but why??
        # self.assertEqual(self.supervisor.unsubscribe_calls, 1)


    async def test_write_data(self):
        db: Session = self.app["db"]
        store: MockStore = self.app['data-store']
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        data = helpers.create_data(stream.layout)
        resp: aiohttp.ClientResponse = await \
            self.client.post("/data", params={"path": "/folder1/stream1"},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.inserted_data)

        # can write stream by id as well
        store.inserted_data = False
        data = helpers.create_data(stream.layout)
        resp: aiohttp.ClientResponse = await \
            self.client.post("/data", params={"id": stream.id},
                             data=data.tobytes())
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.inserted_data)


    async def test_remove_data(self):
        db: Session = self.app["db"]
        store: MockStore = self.app['data-store']
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        resp: aiohttp.ClientResponse = await \
            self.client.delete("/data", params={"path": "/folder1/stream1",
                                                "start": "100", "end": "200"})
        self.assertEqual(resp.status, 200)
        (start, end) = store.removed_data_bounds
        self.assertEqual(start, 100)
        self.assertEqual(end, 200)

        # can remove data by path as well
        resp: aiohttp.ClientResponse = await \
            self.client.delete("/data", params={"id": stream.id,
                                                "start": "800", "end": "900"})
        self.assertEqual(resp.status, 200)
        (start, end) = store.removed_data_bounds
        self.assertEqual(start, 800)
        self.assertEqual(end, 900)


    async def test_interval_data(self):
        db: Session = self.app["db"]
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        resp = await self.client.get("/data/intervals.json",
                                     params={"id": stream.id})
        # response should be a list of intervals
        data = await resp.json()
        self.assertGreater(len(data), 0)
        for interval in data:
            self.assertEqual(2, len(interval))

        # can query by path as well
        resp = await self.client.get("/data/intervals.json",
                                     params={"path": "/folder1/stream1"})
        # response should be a list of intervals
        data = await resp.json()
        self.assertGreater(len(data), 0)
        for interval in data:
            self.assertEqual(2, len(interval))

        # can specify time bounds
        resp = await self.client.get("/data/intervals.json",
                                     params={"path": "/folder1/stream1",
                                             "start": 10,
                                             "end": 20})
        # the mock store returns the start end bounds as the single interval
        data = await resp.json()
        self.assertEqual(data, [[10, 20]])



from tests.api import mock_session
import unittest
import tempfile
import shutil
import os
import asyncio

from joule.api import helpers, TcpNode
from joule import errors

CAFILE = os.path.join(os.path.dirname(__file__), 'ca.joule.crt')


class TestApiHelpers(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        os.environ["JOULE_USER_CONFIG_DIR"] = self.temp_dir.name
        self.ca_path = os.path.join(self.temp_dir.name, "ca.crt")
        shutil.copyfile(CAFILE, self.ca_path)
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.temp_dir.cleanup()
        self.loop.close()

    def test_creates_config_dir(self):
        # require helper to create the directory
        config_dir = os.path.join(self.temp_dir.name, "configs")
        os.environ["JOULE_USER_CONFIG_DIR"] = config_dir
        helpers.get_nodes()
        self.assertTrue(os.path.isdir(config_dir))

    def test_normal_operation(self):
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")
        default_file = os.path.join(self.temp_dir.name, "default_node.txt")

        helpers.save_node(node1)
        helpers.save_node(node2)
        default_node = helpers.get_node()
        # first node should be the default
        self.assertEqual(node1.name, default_node.name)
        # nodes should use the cafile
        self.assertEqual(default_node.session.cafile, self.ca_path)
        # unless it is missing...
        os.remove(self.ca_path)
        retrieved_node2 = helpers.get_node("node2")
        # then nodes do not have a cafile set
        self.assertEqual(retrieved_node2.session.cafile, "")
        self.assertEqual(retrieved_node2.url, node2.url)
        # when the default node is deleted, the next one becomes the default
        helpers.delete_node(node1)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        # if the default file is missing it is automatically created
        os.remove(default_file)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        self.assertTrue(os.path.isfile(default_file))
        # but the default can be changed when the node is added back
        helpers.save_node(node1)
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node2")
        helpers.set_default_node("node1")
        default_node = helpers.get_node()
        self.assertEqual(default_node.name, "node1")
        # all nodes can be retrieved
        nodes = helpers.get_nodes()
        self.assertEqual(len(nodes), 2)
        self.assertTrue(nodes[0].name in ["node1", "node2"])
        self.assertTrue(nodes[1].name in ["node1", "node2"])

    def test_lists_all_nodes(self):
        pass

    def test_delete_node(self):
        # nodes can be deleted by object or name
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")
        helpers.save_node(node1)
        helpers.save_node(node2)
        self.assertEqual(len(helpers.get_nodes()), 2)
        helpers.delete_node(node1)
        helpers.delete_node("node2")

        # removing a non-existent node generates an error
        with self.assertRaises(errors.ApiError):
            helpers.delete_node(node1)


    def test_save_node(self):
        pass



import unittest
import datetime
from joule.models import (DataStream, Element, data_stream)
from joule.errors import ConfigurationError
from tests import helpers


class TestStream(unittest.TestCase):
    def setUp(self):
        us_in_week = int(7 * 24 * 60 * 60 * 1e6)
        self.my_stream = data_stream.DataStream(name="test",
                                                description="test_description",
                                                datatype=DataStream.DATATYPE.FLOAT32,
                                                keep_us=us_in_week,
                                                decimate=True,
                                                updated_at=datetime.datetime.now())
        self.my_stream.elements.append(Element(name="e1", index=0,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS))
        self.base_config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description
                 path = /some/path/to/data
                 datatype = float32
                 keep = 1w
                 decimate = yes
               [Element1]
                 name = e1
            """)

    def test_attributes(self):
        my_stream = DataStream(name='test', description='a test',
                               datatype=DataStream.DATATYPE.FLOAT32, keep_us=DataStream.KEEP_ALL,
                               updated_at=datetime.datetime.now())
        self.assertIsNotNone(my_stream)
        # has a meaningful string representation
        self.assertTrue("test" in "%r" % my_stream)
        # if remote is not set, the pseudo attributes are empty
        self.assertEqual(my_stream.remote_node, '')
        self.assertEqual(my_stream.remote_path, '')
        # if remote is set, these attributes have values
        my_stream.set_remote("remote_url", "remote_path")
        self.assertEqual(my_stream.remote_node, "remote_url")
        self.assertEqual(my_stream.remote_path, "remote_path")

    def test_json(self):
        my_stream = DataStream(id=0, name='test', decimate=True,
                               datatype=DataStream.DATATYPE.UINT16,
                               updated_at=datetime.datetime.now())
        for j in range(4):
            my_stream.elements.append(Element(name="e%d" % j, index=j,
                                              display_type=Element.DISPLAYTYPE.CONTINUOUS))
        # turns streams into json
        json = my_stream.to_json()
        self.assertEqual(json['decimate'], True)
        self.assertEqual(json['name'], 'test')
        self.assertEqual(len(json['elements']), 4)
        # builds streams from json
        new_stream = data_stream.from_json(json)
        self.assertEqual(new_stream.id, my_stream.id)
        self.assertEqual(len(new_stream.elements), len(my_stream.elements))

    def test_parses_config(self):
        s = data_stream.from_config(self.base_config)
        self.assertEqual(s.name, self.base_config["Main"]["name"])

    def test_allows_no_keep(self):
        config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /simple/demo
                 datatype = float32
                 keep = None
               [Element1]
                 name=1
            """)
        s = data_stream.from_config(config)
        self.assertEqual(s.keep_us, 0)

    def test_allows_long_keeps(self):
        config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /simple/demo
                 datatype = float32
                 keep = 100h
               [Element1]
                 name=1
            """)
        s = data_stream.from_config(config)
        self.assertEqual(s.keep_us, 100 * 60 * 60 * 1e6)

    def test_allows_no_description(self):
        self.base_config.remove_option("Main", "description")
        parsed_stream = data_stream.from_config(self.base_config)
        self.assertEqual(parsed_stream.description, "")

    def test_computes_layout(self):
        """data_format returns float32_4 for example"""
        self.my_stream.elements = []
        for i in range(4):
            self.my_stream.elements.append(Element(name="%d" % i))
        self.assertEqual(self.my_stream.layout, "float32_4")
        self.assertEqual(self.my_stream.data_width, 5)

    def test_has_string_representation(self):
        self.assertRegex("%s" % self.my_stream, self.my_stream.name)

    def test_updates(self):
        self.my_stream.update_attributes({
            "name": "new name",
            "description": "new description",
            "elements": [
                {"name": "new name"},
            ]
        })
        self.assertEqual("new name", self.my_stream.name)
        self.assertEqual("new description", self.my_stream.description)
        # updates element attributes
        self.assertEqual("new name", self.my_stream.elements[0].name)

        # number of elements in the request must match the elements in the stream
        with self.assertRaises(ConfigurationError):
            self.my_stream.update_attributes({
                "name": "new name",
                "description": "new description",
                "elements": [
                    {"name": "new name"},
                    {"extra", "causes error"}
                ]
            })

    def test_from_nilmdb_metadata(self):
        metadata = {"name": "sinefit", "name_abbrev": "", "delete_locked": False,
                    "streams": [{"column": 0, "name": "stream_0",
                                 "units": None, "scale_factor": 1.0,
                                 "offset": 0.0, "plottable": False,
                                 "discrete": None, "default_min": -10,
                                 "default_max": None},
                                {"column": 1, "name": "stream_1",
                                 "units": None, "scale_factor": 2.0,
                                 "offset": 3.0, "plottable": True,
                                 "discrete": None, "default_min": None,
                                 "default_max": None},
                                {"column": 2, "name": "stream_2",
                                 "units": None, "scale_factor": 1.0,
                                 "offset": 10.5, "plottable": True,
                                 "discrete": None, "default_min": None,
                                 "default_max": None}]}
        my_stream = data_stream.from_nilmdb_metadata(metadata, "float32_3")
        # make sure the stream is created correctly
        self.assertEqual(my_stream.name, "sinefit")
        self.assertEqual(my_stream.layout, "float32_3")
        self.assertEqual(len(my_stream.elements), 3)
        my_stream.elements.sort(key=lambda e: e.index)
        elem0 = my_stream.elements[0]
        self.assertEqual(elem0.index, 0)
        self.assertEqual(elem0.name, 'stream_0')
        self.assertEqual(elem0.plottable, False)
        self.assertEqual(elem0.default_min, -10)
        self.assertEqual(elem0.default_max, None)
        elem2 = my_stream.elements[2]
        self.assertEqual(elem2.index, 2)
        self.assertEqual(elem2.name, 'stream_2')
        self.assertEqual(elem2.plottable, True)
        self.assertEqual(elem2.scale_factor, 1.0)
        self.assertEqual(elem2.offset, 10.5)

    def test_from_nilmdb_no_metadata(self):
        my_stream = data_stream.from_nilmdb_metadata({'name': 'test'}, "float32_3")
        # make sure the stream is created correctly
        self.assertEqual(my_stream.name, "test")
        self.assertEqual(my_stream.layout, "float32_3")
        self.assertEqual(len(my_stream.elements), 3)
        my_stream.elements.sort(key=lambda e: e.index)
        elem0 = my_stream.elements[0]
        self.assertEqual(elem0.index, 0)
        self.assertEqual(elem0.name, 'Element1')
        self.assertEqual(elem0.plottable, True)
        self.assertEqual(elem0.default_min, None)
        self.assertEqual(elem0.default_max, None)
        elem2 = my_stream.elements[2]
        self.assertEqual(elem2.index, 2)
        self.assertEqual(elem2.name, 'Element3')
        self.assertEqual(elem2.plottable, True)
        self.assertEqual(elem2.scale_factor, 1.0)
        self.assertEqual(elem2.offset, 0.0)

    def test_to_nilmdb_metadata(self):
        my_stream = DataStream(name="test", datatype=DataStream.DATATYPE.INT16,
                               updated_at=datetime.datetime.now())
        my_stream.elements.append(
            Element(name="e0", index=0, plottable=True, offset=0.0, default_min=-5, scale_factor=1.0))
        my_stream.elements.append(
            Element(name="e1", index=1, plottable=True, units="watts", offset=3.5, scale_factor=1.0))
        metadata = my_stream.to_nilmdb_metadata()
        # NOTE: The plottable flag is always True (see TODO in element.py)
        expected = {"name": "test", "name_abbrev": "", "delete_locked": False,
                    "streams": [{"column": 0, "name": "e0",
                                 "units": None, "scale_factor": 1.0,
                                 "offset": 0.0, "plottable": True,
                                 "discrete": False, "default_min": -5,
                                 "default_max": None},
                                {"column": 1, "name": "e1",
                                 "units": "watts", "scale_factor": 1.0,
                                 "offset": 3.5, "plottable": True,
                                 "discrete": False, "default_min": None,
                                 "default_max": None}]}
        self.assertEqual(expected, metadata)



#!/usr/bin/python3 -u

"""
Run the configuration in follower
"""

import os
import sys
import subprocess
import time
import shlex
import socket
import signal

SOURCE_DIR = "/joule"
MODULE_SCRIPT_DIR = "/joule/tests/e2e/module_scripts"
JOULE_CONF_DIR = "/joule/tests/e2e/follower"


def prep_system():
    os.symlink(MODULE_SCRIPT_DIR, "/module_scripts")


def run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL):
    return subprocess.run(shlex.split(cmd), stdout=stdout, stderr=stderr)


def main():
    prep_system()

    config_file = os.path.join(JOULE_CONF_DIR, "main.conf")
    subprocess.run(("joule admin erase --yes --config %s" % config_file).split(" "))

    # add a key entry for the master node
    # get API permissions
    os.environ["LOGNAME"] = "e2e"
    os.environ["JOULE_USER_CONFIG_DIR"] = "/tmp/joule_user"
    with open(os.devnull, 'w') as devnull:
        subprocess.run(("joule admin authorize --config %s" % config_file).split(" "),
                       stdout=devnull)

    jouled = subprocess.Popen(["jouled", "--config", config_file],
                              stdout=sys.stdout,
                              stderr=sys.stderr,
                              universal_newlines=True)
    # wait until local node is online
    max_tries = 16
    num_tries = 0
    while num_tries < max_tries:
        num_tries += 1
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        result = sock.connect_ex(('localhost', 8088))
        sock.close()
        if result == 0:
            break
        time.sleep(0.5)
    # local node failure, print the log
    if num_tries == max_tries:
        exit()
    print("follower node is up!")
    sys.stdout.flush()
    # wait until the master node is online
    max_tries = 10
    num_tries = 0
    while num_tries < max_tries:
        num_tries += 1
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex(('node1.joule', 8088))
        sock.close()
        if result == 0:
            break
        else:
            print("cannot find node1.joule %d" % result)
        time.sleep(0.5)
    # remote node failure, print the log
    if num_tries == max_tries:
        print("Cannot find node1.joule, exiting follower")
        exit()

    # follow node1.joule
    #    with open(os.devnull, 'w') as devnull:
    subprocess.run("joule master add joule https://node1.joule:8088".split(" "))  # ,
    #                       #stderr=devnull, stdout=devnull)
    # subprocess.run(f"jouled --config {config_file})")
    # this will just hang, node1 exits and terminates this container
    stdout, _ = jouled.communicate()
    print(stdout)
    print(_)
    return 0


if __name__ == "__main__":
    exit(main())



from click.testing import CliRunner

import warnings
import numpy as np
import logging
import tempfile
import h5py
import json
import datetime
import unittest
from icecream import ic
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main
from joule.models import DataStream, Element, StreamInfo
from joule.api.data_stream import Element as ApiElement
from tests import helpers

warnings.simplefilter('always')
log = logging.getLogger('aiohttp.access')
log.setLevel(logging.WARNING)


class TestDataIngest(FakeJouleTestCase):

    def test_ingests_data_to_empty_existing_stream(self):
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="existing", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.datetime.utcnow())
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        src_data = helpers.create_data(src.layout, length=22000)
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/existing', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data)
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name, '--stream', '/test/existing'])
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
        db_obj = self.msgs.get()
        np.testing.assert_array_equal(src_data, db_obj.data)
        # uses the stream parameter instead of the hd5 attrs
        self.assertEqual(db_obj.stream.name, 'existing')
        self.stop_server()

    def test_datatype_mismatch(self):
        #  the datatype of the file and target stream must match
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.UINT16, updated_at=datetime.datetime.utcnow())
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        file_data = helpers.create_data('int16_3')
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/dest', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, file_data)
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name])
            self.assertIn("datatype", result.output)
            self.assertNotEqual(result.exit_code, 0)

        self.stop_server()

    def test_confirms_data_removal(self):
        # if there is existing data in the target, confirm removal
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.datetime.utcnow())
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        # source has 100 rows of data between [0, 100]
        src_data = helpers.create_data(src.layout, length=1000)
        # File:    |----|
        # DataStream:     |-----|
        src_info = StreamInfo(int(src_data['timestamp'][500]),
                              int(src_data['timestamp'][-1]),
                              500)
        server.add_stream('/test/dest', src, src_info, src_data[500:])
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data[:750])
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name],
                                   input='N')
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            self.assertIn('Cancelled', result.output)

        self.stop_server()

    def test_dataset_mismatch(self):
        # the length of [data] and [timestamp]  must match
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            with h5py.File(data_file.name, 'w') as hdf_root:
                hdf_root.create_dataset('data', (10, 10),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')
                hdf_root.create_dataset('timestamp', (11, 1),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)

    def test_stream_element_mismatch(self):
        #  the datatype of the file and target stream must match
        server = FakeJoule()
        # create the source stream
        src = DataStream(id=0, name="dest", keep_us=100,
                         datatype=DataStream.DATATYPE.UINT16, updated_at=datetime.datetime.utcnow())
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(4)]
        # source has 100 rows of data between [0, 100]
        file_data = helpers.create_data('uint16_3')
        src_info = StreamInfo(0, 0, 0, 0)
        server.add_stream('/test/dest', src, src_info, None)
        self.start_server(server)
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, file_data)
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name])
            self.assertIn("elements", result.output)
            self.assertNotEqual(result.exit_code, 0)

        self.stop_server()

    def test_creates_stream_from_attributes(self):
        # creates the target stream based on hdf attrs
        server = FakeJoule()
        src_data = helpers.create_data('float32_3')
        self.start_server(server)
        runner = CliRunner()

        with tempfile.NamedTemporaryFile() as data_file:
            write_hd5_data(data_file, src_data[:750])
            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name],
                                   input='N')
            _print_result_on_error(result)
            self.assertEqual(result.exit_code, 0)
            # prints a message about the new stream
            self.assertIn('/test/dest', result.output)
        db_entry = self.msgs.get()
        # make sure all the elements are present
        elem_names = [e.name for e in db_entry.stream.elements]
        expected_names = ['Test %d' % i for i in range(3)]
        self.assertEqual(elem_names, expected_names)
        self.assertEqual(db_entry.stream.name, 'dest')
        self.stop_server()

    def test_invalid_data_structure(self):
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as data_file:
            with h5py.File(data_file.name, 'w') as hdf_root:
                hdf_root.create_dataset('data', (1, 10),
                                        dtype=np.dtype('float32'),
                                        compression='gzip')

            result = runner.invoke(main, ['data', 'ingest', '--file', data_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)

    def test_invalid_filetype(self):
        runner = CliRunner()
        with tempfile.NamedTemporaryFile() as bad_file:
            with open(bad_file.name, 'w') as f:
                f.write('invalid filetype')
            result = runner.invoke(main, ['data', 'ingest', '--file', bad_file.name])
            self.assertIn("Error", result.output)
            self.assertNotEqual(result.exit_code, 0)


def write_hd5_data(path, data):
    with h5py.File(path, 'w') as hdf_root:
        hdf_data = hdf_root.create_dataset('data', data['data'].shape,
                                           dtype=data.dtype[1].base,
                                           compression='gzip')
        hdf_timestamps = hdf_root.create_dataset('timestamp', [len(data), 1],
                                                 dtype='i8',
                                                 compression='gzip')
        hdf_data[...] = data['data']
        hdf_timestamps[...] = data['timestamp'][:, None]
        hdf_root.attrs['path'] = '/test/dest'
        elements = [ApiElement(name='Test %d' % i,
                               display_type='discrete') for i in range(data['data'].shape[1])]
        element_json = json.dumps([e.to_json() for e in elements])
        hdf_root.attrs['element_json'] = element_json


def _print_result_on_error(result):
    if result.exit_code != 0:
        print("output: ", result.output)
        print("exception: ", result.exception)



import os
import argparse
import asyncio
import json
import numpy as np
from unittest import mock

from .test_filter_module import SimpleFilter
from .test_reader_module import SimpleReader
from joule.client import CompositeModule
from joule.api import BaseNode, DataStream
from joule.models import DataStream, Element, pipes
from joule import api
from tests import helpers
import warnings

warnings.simplefilter('always')


class SimpleComposite(CompositeModule):
    # runs the simple reader and simple filter
    # as a composite module
    async def setup(self, parsed_args, inputs, outputs):
        reader_module = SimpleReader()
        filter_module = SimpleFilter()

        pipe = pipes.LocalPipe("float32_3")
        reader_task = reader_module.run(parsed_args, pipe)
        filter_task = filter_module.run(parsed_args, {'to_filter': pipe},
                                        {'from_filter': outputs['output']})
        return [reader_task, filter_task]


class TestCompositeModule(helpers.AsyncTestCase):

    def setUp(self):
        super().setUp()
        # module output is a float32_3 stream
        self.stream = DataStream(name="output", datatype=DataStream.DATATYPE.FLOAT32,
                                 elements=[Element(name="e%d" % j, index=j,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])

    def test_runs_composited_modules(self):
        module = SimpleComposite()
        (r, w) = os.pipe()
        rf = pipes.reader_factory(r)
        pipe = pipes.InputPipe(name="output", stream=self.stream, reader_factory=rf)
        pipe_arg = json.dumps(json.dumps({"outputs": {'output': {'fd': w, 'id': None, 'layout': self.stream.layout}},
                                          "inputs": {}}))
        data = helpers.create_data(self.stream.layout)
        args = argparse.Namespace(pipes=pipe_arg, socket="unset",
                                  node="", api_socket="",
                                  url='http://localhost:8080',
                                  mock_data=data)
        # run the composite module
        module.start(args)
        asyncio.set_event_loop(self.loop)
        # check the output
        received_data = asyncio.run(pipe.read())
        np.testing.assert_array_equal(data['timestamp'], received_data['timestamp'])
        np.testing.assert_array_almost_equal(data['data'] * 2, received_data['data'])
        asyncio.run(pipe.close())

    def test_handles_bad_pipe_configs(self):
        args = argparse.Namespace(pipes="invalid config", socket="unset",
                                  node="", api_socket="",
                                  url='http://localhost:8080')
        module = SimpleComposite()
        with self.assertLogs(level="ERROR"):
            module.start(args)
        asyncio.set_event_loop(self.loop)

